#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_dict_arithmetic.py.

Contains the DictArithmetic class. See ``help(qubovert.utils.DictArithmetic)``.
It is used to define operations for dictionaries, and is a parent class to some
other classes, such as ``qubovert.utils.QUBOMatrix`` and ``qubovert.utils.BO``.

"""

from . import subgraph, subvalue


__all__ = 'DictArithmetic',


def _generate_key_value_pairs(*args, **kwargs):
    """_generate_key_value_pairs.

    Generate the key, value pairs from the parameter inputs to a dictionary
    without explicitly recreating the dictionary.

    Parameters
    ----------
    arguments : define a dictionary.
        args and kwargs define a dictionary. See ``help(dict)``.

    Yields
    ------
    pairs : tuple (key, value).

    """
    if not kwargs and len(args) == 1:
        a = args[0]
        k_v_pairs = (x for x in (a.items() if isinstance(a, dict) else a))
    elif not args and kwargs:
        k_v_pairs = (x for x in kwargs.items())
    else:
        k_v_pairs = (x for x in dict(*args, **kwargs).items())

    yield from k_v_pairs


class DictArithmetic(dict):
    """DictArithmetic.

    A class to handle dictionaries. It is the same thing as a dictionary
    with some methods modified.

    One method is that values will always default to 0. Consider the following
    example:

    >>> d = DictArithmetic()
    >>> print(d[(0, 0)]) # will print 0
    >>> d[(0, 0)] += 1
    >>> print(d) # will print {(0, 0): 1}

    Compared to an ordinary dictionary.

    >>> g = dict()
    >>> print(g[(0, 0)]) # will raise KeyError
    >>> g[(0, 0)] += 1 # will raise KeyError, since (0, 0) was never set

    One method is that if we set an item to 0, it will be removed. Consider
    the following example:

    >>> d = DictArithmetic()
    >>> d[(0, 0)] += 1
    >>> d[(0, 0)] -= 1
    >>> print(d) # will print {}

    One method is that if we initialize DictArithmetic with a previous
    dictionary it will be reinitialized to ensure that the DictArithmetic
    contains no zero values. Consider the following example:

    >>> d = DictArithmetic({0: 1, (1, 0): 2, (2, 0): 0})
    >>> print(d) # will print {0: 1, (1, 0): 2}

    We also change the update method so that it follows all the conventions.

    >>> d = DictArithmetic({'a': 1, (0, 1): 2})
    >>> d.update({'a': 0, (1, 1): -1})
    >>> print(d)  # will print {(0, 1): 2, (1, 1): -1}

    We also include arithmetic, addition, subtraction, scalar division,
    scalar multiplication, and all those in place. For example,

    >>> d = DictArithmetic({(0, 0): 1, (0, 1): -2})
    >>> g = d + {(0, 0): -1}
    >>> print(g) # will print {(0, 1): -2}
    >>> g *= 4
    >>> print(g) # will print {(0, 1): -8}
    >>> g -= {(0, 1): -8}
    >>> print(g) # will print {}

    Adding or subtracting constants will update the () element of the
    dict.

    >>> d = DictArithmetic()
    >>> d += 5
    >>> print(d)
    {(): 5}

    You can give it a name.

    >>> d = DictArithmetic()
    >>> d.name
    None
    >>> d.name = 'd'
    >>> d.name
    'd'

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        Initialize the object. If you supply args and kwargs that
        represent a dictionary, they will be reinitialized to follow the
        conventions set in ``__setitem__``.

        Parameters
        ----------
        arguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class.

        """
        super().__init__()
        # reset to make sure everything is in the proper form

        for key, value in _generate_key_value_pairs(*args, **kwargs):
            self[key] += value

        self.name = None

    @property
    def name(self):
        """name.

        Return the name of the object.

        Return
        ------
        name : object.

        Example
        -------
        >>> d = DictArithmetic()
        >>> d.name
        None
        >>> d.name = 'd'
        >>> d.name
        'd'

        """
        return self._name

    @name.setter
    def name(self, name):
        """name.

        Set the name of the object.

        Parameters
        ----------
        name : object.

        Example
        -------
        >>> d = DictArithmetic()
        >>> d.name
        None
        >>> d.name = 'd'
        >>> d.name
        'd'

        """
        self._name = name

    def __getitem__(self, key):
        """__getitem__.

        Overrides the dict.__getitem__ command so that a KeyError is not
        thrown if `key` is not in the dictionary. If the key is not present,
        then we return 0. See the ``dict.__getitem__`` docstring.

        """
        return self.get(key, 0)

    def __setitem__(self, key, value):
        """__setitem__.

        Overrides the dict.__setitem__ command. If `value` is equal to 0, then
        the key will be removed from the dictionary. See the
        ``dict.__setitem__`` for more.

        """
        if value:
            super().__setitem__(key, value)
        else:
            self.pop(key, 0)

    @property
    def num_terms(self):
        """num_terms.

        Return the number of terms in the dictionary.

        Return
        ------
        n : int.
            Number of terms in the dictionary.

        """
        return len(self)

    def copy(self):
        """copy.

        Same as dict.copy, but we adjust the method so that it returns a
        DictArithmetic object, or whatever object is the subclass.

        Returns
        -------
        d : DictArithmetic object, or subclass of.
            Same as ``self.__class__``.

        """
        return self.__class__(self)

    def update(self, *args, **kwargs):
        """update.

        Update the dictionary but following all the conventions of this class.

        Parameters
        ----------
        arguments : defines a dictionary, ie ``d = dict(*args, **kwargs)``.
            Each element in d will be added in place to this instance following
            all the required convensions.

        """
        for k, v in _generate_key_value_pairs(*args, **kwargs):
            self[k] = v

    def __add__(self, other):
        """__add__.

        Add two DictArithmetics or dicts, return the sum.

        Parameters
        ----------
        other : a DictArithmetic or dict object.

        Return
        ------
        d : a DictArithmetic object, self + other.

        Examples
        --------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0, 0): 2, (0, 1): 1}
        >>> print(d + g)
        {(0, 0): 3}

        """
        d = self.copy()
        d += other
        return d

    def __radd__(self, other):
        """__radd__.

        Add two DictArithmtic objects or dicts.

        Parameters
        ----------
        other : a DictArithmetic or dict object.

        Return
        ------
        d : a DictArithmetic object, self + other.

        Examples
        --------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0, 0): 2, (0, 1): 1}
        >>> print(g + d)
        {(0, 0): 3}

        """
        return self + other

    def __iadd__(self, other):
        """__iadd__.

        Same as the __add__ method, but done in place.

        Parameters
        ----------
        other : a DictArithmetic or dict object, or number.

        Return
        ------
        d : a DictArithmetic object, self.

        Examples
        --------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0, 0): 2, (0, 1): 1}
        >>> d += g
        >>> print(d)
        {(0, 0): 3}

        """
        if isinstance(other, dict):
            for k, v in other.items():
                self[k] += v
        else:
            self[()] += other
        return self

    def __sub__(self, other):
        """__sub__.

        Subtract two DictArithmetic or dicts, return the difference.

        Parameters
        ----------
        other : a DictArithetic or dict object.

        Return
        ------
        d : a DictArithmetic object, self - other.

        Examples
        --------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0, 0): 2, (0, 1): 1}
        >>> print(d - g)  # will print {(0, 0): -1}

        """
        d = self.copy()
        d -= other
        return d

    def __rsub__(self, other):
        """__rsub__.

        Subtract two DictArithmetic or dicts.

        Parameters
        ----------
        other : a DictArithmetic or dict object.

        Return
        ------
        d : a DictArithmetic object, other - self.

        Examples
        --------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0, 0): 2, (0, 1): 1}
        >>> print(g - d)  # will print {(0, 0): 1}

        """
        return -1*self + other

    def __isub__(self, other):
        """__isub__.

        Same as the __sub__ method, but done in place.

        Parameters
        ----------
        other : a DictArithmetic or dict object.

        Returns
        -------
        d : a DictArithmetic object, self.

        Examples
        --------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0, 0): 2, (0, 1): 1}
        >>> d -= g
        >>> print(d)  # will print {(0, 0): -1}

        """
        if isinstance(other, dict):
            for k, v in other.items():
                self[k] -= v
        else:
            self[()] -= other
        return self

    def __mul__(self, other):
        """__mul__.

        Multiplying a DictArithmetic by a scalar or another dict.

        Parameters
        ----------
        other : numeric or dict/DictArithmetic object.

        Return
        ------
        d : a DictArithmetic object.

        Example
        -------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> print(d * 2)
        {(0, 0): 2, (0, 1): -1}

        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0,): -1, (0, 2): 1}
        >>> print(d * g)
        {(0, 0, 0): -1, (0, 0, 0, 2): 1, (0, 1, 0): 1, (0, 1, 0, 2): -1}

        Note that if the keys are not tuples, then they will be made tuples! Ie

        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {0: -1, 2: 1}
        >>> print(d * g)
        {(0, 0, 0): -1, (0, 0, 2): 1, (0, 1, 0): 1, (0, 1, 2): -1}

        """
        d = self.copy()
        d *= other
        return d

    def __rmul__(self, other):
        """__rmul__.

        Same as __mul__, but for different ordering.

        Parameters
        ----------
        other : numeric or dict/DictArithmetic object.

        Return
        ------
        d : a DictArithmetic object.

        Example
        -------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> print(2 * d)
        {(0, 0): 2, (0, 1): -1}

        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0,): -1, (0, 2): 1}
        >>> print(g * d)
        {(0, 0, 0): -1, (0, 0, 0, 2): 1, (0, 1, 0): 1, (0, 1, 0, 2): -1}

        Note that if the keys are not tuples, then they will be made tuples! Ie

        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {0: -1, 2: 1}
        >>> print(g * d)
        {(0, 0, 0): -1, (0, 0, 2): 1, (0, 1, 0): 1, (0, 1, 2): -1}

        """
        return self * other

    def __imul__(self, other):
        """__imul__.

        Same as __mul__, but done in place.

        Parameters
        ----------
        other : numeric or dict/DictArithmetic object.

        Return
        ------
        d : a DictArithmetic object, self.

        Example
        -------
        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> d *= 2
        >>> print(d)
        {(0, 0): 2, (0, 1): -1}

        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {(0,): -1, (0, 2): 1}
        >>> d *= g
        >>> print(d)
        {(0, 0, 0): -1, (0, 0, 0, 2): 1, (0, 1, 0): 1, (0, 1, 0, 2): -1}

        Note that if the keys are not tuples, then they will be made tuples! Ie

        >>> d = DictArithmetic({(0, 0): 1, (0, 1): -1})
        >>> g = {0: -1, 2: 1}
        >>> d *= g
        >>> print(d)
        {(0, 0, 0): -1, (0, 0, 2): 1, (0, 1, 0): 1, (0, 1, 2): -1}

        """
        if isinstance(other, dict):
            items, oitems = tuple(self.items()), tuple(other.items())
            self.clear()
            for k, v in items:
                kp = k if isinstance(k, tuple) else (k,)
                for ko, vo in oitems:
                    kop = ko if isinstance(ko, tuple) else (ko,)
                    self[kp + kop] += v * vo

        else:
            for k in tuple(self.keys()):
                self[k] *= other

        return self

    def __pow__(self, exponent):
        """__pow__.

        Raise the object to an integer power. Note that for example
        ``self ** 3 == self * self * self``.

        Parameters
        ----------
        exponent : int.
            Integer power to raise the DictArithmetic to.

        Returns
        -------
        res : DictArithmetic.

        """
        d = self.copy()
        d **= exponent
        return d

    def __ipow__(self, exponent):
        """__ipow__.

        Same as ``__pow__`` but in place.

        Raise the object to an integer power. Note that for example
        ``self ** 3 == self * self * self``.

        Parameters
        ----------
        exponent : int.
            Integer power to raise the DictArithmetic to.

        Returns
        -------
        self : DictArithmetic.

        """
        if not isinstance(exponent, int) or exponent <= 0:
            raise ValueError("Exponent must be a positive integer")

        if exponent > 1:
            old = self.copy()
            for _ in range(exponent-1):
                self *= old

        return self

    def __truediv__(self, other):
        """__truediv__.

        Dividing a DictArithmetic by a scalar.

        Parameters
        ----------
        other : numeric.

        Return
        ------
        d : a DictArithmetic object.

        Example
        -------
        >>> d = DictArithmetic((0, 0)=1, (0, 1)=-1)
        >>> print(d / 2)
        {(0, 0): .5, (0, 1): -.5}

        """
        d = self.copy()
        d /= other
        return d

    def __itruediv__(self, other):
        """__itruediv__.

        Same as __truediv__, but done in place.

        Parameters
        ----------
        other : numeric.

        Return
        ------
        d : a DictArithmetic object, self.

        Example
        -------
        >>> d = DictArithmetic((0, 0)=1, (0, 1)=-1)
        >>> d /= 2
        >>> print(d)
        {(0, 0): .5, (0, 1): -.5}

        """
        for k in tuple(self.keys()):
            self[k] /= other
        return self

    def __floordiv__(self, other):
        """__floordiv__.

        Floor dividing a DictArithmetic by a scalar.

        Parameters
        ----------
        other : numeric.

        Return
        ------
        d : a DictArithmetic object.

        Example
        -------
        >>> d = DictArithmetic((0, 0)=2, (0, 1)=-1)
        >>> print(d // 2)
        {(0, 0): 1, (0, 1): 0}

        """
        d = self.copy()
        d //= other
        return d

    def __ifloordiv__(self, other):
        """__ifloordiv__.

        Same as __floordiv__, but done in place.

        Parameters
        ----------
        other : numeric.

        Return
        ------
        d : a DictArithmetic object, self.

        Example
        -------
        >>> d = DictArithmetic((0, 0)=2, (0, 1)=-1)
        >>> d //= 2
        >>> print(d)
        {(0, 0): 1, (0, 1): 0}

        """
        for k in tuple(self.keys()):
            self[k] //= other
        return self

    def __pos__(self):
        """__pos__.

        Called when unary + is used. Returns copy of self.

        Returns
        -------
        res : DictArithmetic object.
            Copy of self.

        """
        return self.copy()

    def __neg__(self):
        """__neg__.

        Called when unary - is used. Returns -1 * self.

        Returns
        -------
        res : DictArithmetic object.
            -1 * self.

        """
        return -1 * self

    def normalize(self, value=1):
        """normalize.

        Normalize the coefficients to a maximum magnitude.

        Parameters
        ----------
        value : float (optional, defaults to 1).
            Every coefficient value will be normalized such that the
            coefficient with the maximum magnitude will be +/- 1.

        Examples
        --------
        >>> from qubovert.utils import DictArithmetic
        >>> d = DictArithmetic({(0, 1): 1, (1, 2, 'x'): 4})
        >>> d.normalize()
        >>> print(d)
        {(0, 1): 0.25, (1, 2, 'x'): 1}

        >>> from qubovert.utils import DictArithmetic
        >>> d = DictArithmetic({(0, 1): 1, (1, 2, 'x'): -4})
        >>> d.normalize()
        >>> print(d)
        {(0, 1): 0.25, (1, 2, 'x'): -1}

        >>> from qubovert import PUBO
        >>> d = PUBO({(0, 1): 1, (1, 2, 'x'): 4})
        >>> d.normalize()
        >>> print(d)
        {(0, 1): 0.25, (1, 2, 'x'): 1}

        >>> from qubovert.utils import PUBO
        >>> d = PUBO({(0, 1): 1, (1, 2, 'x'): -4})
        >>> d.normalize()
        >>> print(d)
        {(0, 1): 0.25, (1, 2, 'x'): -1}

        """
        mult = value / max(abs(v) for v in self.values())
        for k in self:
            self[k] *= mult

    def subgraph(self, nodes, connections=None):
        """subgraph.

        Create the subgraph of ``self`` that only includes vertices in
        ``nodes``, and external nodes are given the values in ``connections``.

        Parameters
        ----------
        nodes : set.
            Nodes of ``self`` to include in the subgraph.
        connections : dict (optional, defaults to {}).
            For each node in ``self`` that is not in ``nodes``, we assign a
            value given by ``connections.get(node, 0)``.

        Return
        ------
        D : same as type(self).
            The subgraph of ``self`` with nodes in ``nodes`` and the values of
            the nodes not included given by ``connections``.

        Notes
        -----
        Any offset value included in ``self`` (ie {(): 1}) will be ignored,
        however there may be an offset in the output ``D``.

        Examples
        --------
        >>> G = DictArithmetic(
        >>>     {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
        >>> )
        >>> D = G.subgraph({0, 2}, {1: 5})
        >>> D
        {(0,): -17, (0, 2): -1, (): 10}

        >>> G = DictArithmetic(
        >>>     {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
        >>> )
        >>> D = G.subgraph({0, 2})
        >>> D
        {(0, 2): -1, (0,): 3}

        >>> G = DictArithmetic(
        >>>     {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
        >>> )
        >>> D = G.subgraph({0, 1}, {2: -10})
        >>> D
        {(0, 1): -4, (0,): 13, (1,): 2}

        >>> G = DictArithmetic(
        >>>     {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
        >>> )
        >>> D = G.subgraph({0, 1})
        >>> D
        {(0, 1): -4, (0,): 3, (1,): 2}

        """
        return subgraph(self, nodes, connections)

    def subvalue(self, values):
        """subvalue.

        Replace each element in ``self`` with a value in ``values`` if it
        exists.

        Parameters
        ----------
        values : dict.
            For each node ``v`` in ``self`` that is in ``values``, we replace
            the node with ``values[v]``.

        Return
        ------
        D : same as type(self).

        Examples
        --------
        >>> G = DictArithmetic(
        >>>     {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2
        >>> }
        >>> D = G.subvalue({0: 2})
        >>> D
        {(1,): -6, (2,): -2, (): 8}

        >>> G = DictArtihmetic(
        >>>     {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2
        >>> }
        >>> D = G.subvalue({2: -3})
        >>> D
        {(0, 1): -4, (0,): 6, (1,): 2, (): 2}

        >>> G = PUBO(
        >>>     {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2
        >>> }
        >>> D = G.subvalue({2: -3})
        >>> D
        {(0, 1): -4, (0,): 6, (1,): 2, (): 2}

        """
        return subvalue(values, self)

    def __round__(self, ndigits=None):
        """round.

        Round values of the DictArithmetic object.

        Parameters
        ----------
        ndigits : int.
            Number of decimal digits to round to.

        Returns
        -------
        res : DictArithmetic object.
            Copy of self but with each value rounded to ``ndigits`` decimal
            digits. Each value has a type according to the docstring
            specifications of ``round``, see ``help(round)``.

        """
        d = self.__class__()
        for k, v in self.items():
            try:
                d[k] = round(v, ndigits)
            except TypeError:  # symbols don't have round methods
                pass
        return d

    def subs(self, *args, **kwargs):
        """subs.

        Replace any ``sympy`` symbols that are used in the dict with values.
        Please see ``help(sympy.Symbol.subs)`` for more info.

        Parameters
        ----------
        arguments : substitutions.
            Same parameters as are inputted into ``sympy.Symbol.subs``.

        Returns
        -------
        res : DictArithmetic object.
            Same as ``self`` but with all the symbols replaced with values.

        """
        d = self.__class__()

        for k, v in self.items():
            try:
                val = float(v.subs(*args, **kwargs))
            except AttributeError:
                val = v
            except TypeError:
                val = v.subs(*args, **kwargs)
            finally:
                d[k] = val

        return d

    def simplify(self):
        """simplify.

        If ``self`` has any symbolic expressions, this will go through and
        simplify them. This will also make everything a float!

        Return
        ------
        None. Updates it in place.

        """
        # it might be changing size, so convert items to a tuple
        for k, v in tuple(self.items()):
            try:
                self[k] = v.simplify() * 1.  # make everything a float
            except AttributeError:
                self[k] *= 1.  # make it a float



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``sat`` is a module for converting SAT problems into QUBO/QUSO form.

``sat`` is a library of ``qubovert`` for converting satisfiability problems
into PUBOs (see ``help(qubovert.PUBO)``).

"""

from ._satisfiability import *

from ._satisfiability import __all__ as __all_sat__

__all__ = (
    __all_sat__
)

name = "sat"



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_solve_bruteforce.py.

This file contains bruteforce solvers for QUBO/PUBO and QUSO/PUSO, as well
as QUBO/PUBO and PUSO objective function evaluators.

"""

__all__ = (
    'boolean_to_spin', 'spin_to_boolean',
    'decimal_to_spin', 'spin_to_decimal',
    'decimal_to_boolean', 'boolean_to_decimal',
    'pubo_value', 'qubo_value', 'puso_value', 'quso_value',
    'solve_pubo_bruteforce', 'solve_qubo_bruteforce',
    'solve_puso_bruteforce', 'solve_quso_bruteforce'
)


def boolean_to_spin(x):
    """boolean_to_spin.

    Convert a boolean number in {0, 1} to a spin in {1, -1}, in that order.

    Parameters
    ----------
    x : int, iterable of ints, or dict mapping labels to ints.
        Each integer is either 0 or 1.

    Returns
    -------
    z : int, iterable of ints, or dict mapping labels to ints.
        Each integer is either 1 or -1.

    Example
    -------
    >>> boolean_to_spin(0)  # will print 1
    >>> boolean_to_spin(1)  # will print -1
    >>> boolean_to_spin([0, 1, 1])  # will print [1, -1, -1]
    >>> boolean_to_spin({"a": 0, "b": 1})  # will print {"a": 1, "b": -1}

    """
    convert = {0: 1, 1: -1}
    if isinstance(x, (int, float)) and x in convert:
        return convert[x]
    elif isinstance(x, dict):
        return {k: convert[v] for k, v in x.items()}
    return type(x)(convert[i] for i in x)


def spin_to_boolean(z):
    """spin_to_boolean.

    Convert a spin in {1, -1} to a boolean variable in {0, 1}, in that order.

    Parameters
    ----------
    z : int, iterable of ints, or dict mapping labels to ints.
        Each integer is either 1 or -1.

    Returns
    -------
    x : int, iterable of ints, or dict mapping labels to ints.
        Each integer is either 0 or 1.

    Example
    -------
    >>> spin_to_boolean(-1)  # will print 1
    >>> spin_to_boolean(1)  # will print 0
    >>> spin_to_boolean([-1, 1, 1])  # will print [1, 0, 0]
    >>> spin_to_boolean({"a": -1, "b": 1})  # will print {"a": 1, "b": 0}

    """
    convert = {-1: 1, 1: 0}
    if isinstance(z, (int, float)) and z in convert:
        return convert[z]
    elif isinstance(z, dict):
        return {k: convert[v] for k, v in z.items()}
    return type(z)(convert[i] for i in z)


def decimal_to_boolean(d, num_bits=None):
    """decimal_to_boolean.

    Convert the integer ``d`` to its boolean representation.

    Parameters
    ----------
    d : int >= 0.
        Number to convert to binary.
    num_bits : int >= 0 (optional, defaults to None).
        Number of bits in the representation. If ``num_bits is None``, then
        the minimum number of bits required will be used.

    Returns
    -------
    b : tuple of length ``num_bits``.
        Each element of ``b`` is a 0 or 1.

    Example
    -------
    >>> decimal_to_boolean(10, 7)
    (0, 0, 0, 1, 0, 1, 0)

    >>> decimal_to_boolean(10)
    (1, 0, 1, 0)

    """
    if int(d) != d or d < 0:
        raise ValueError("``d`` must be an integer >- 0.")
    b = bin(d)[2:]
    lb = len(b)
    if num_bits is None:
        num_bits = lb
    elif num_bits < lb:
        raise ValueError("Not enough bits to represent the number.")
    return (0,) * (num_bits - lb) + tuple(int(x) for x in b)


def boolean_to_decimal(b):
    """boolean_to_decimal.

    Convert a bit string to its decimal form.

    Parameters
    ----------
    b : tuple or list of 0s and 1s.
        The binary bit string.

    Returns
    -------
    d : int.

    Examples
    --------
    >>> boolean_to_decimal((1, 1, 0))
    6

    """
    return int("".join(str(x) for x in b), base=2) if b else 0


def decimal_to_spin(d, num_spins=None):
    """decimal_to_spin.

    Convert the integer ``d`` to its spin representation (ie its binary
    representation, but with 1 and -1 instead of 0 and 1).

    Parameters
    ----------
    d : int >= 0.
        Number to convert to binary.
    num_spins : int >= 0 (optional, defaults to None).
        Number of bits in the representation. If ``num_spins is None``, then
        the minimum number of bits required will be used.

    Returns
    -------
    b : tuple of length ``num_spins``.
        Each element of ``b`` is a 0 or 1.

    Example
    -------
    >>> decimal_to_spin(10, 7)
    (1, 1, 1, -1, 1, -1, 1)

    >>> decimal_to_spin(10)
    (-1, 1, -1, 1)

    """
    return boolean_to_spin(decimal_to_boolean(d, num_spins))


def spin_to_decimal(b):
    """spin_to_decimal.

    Convert a spin string to its decimal form.

    Parameters
    ----------
    b : tuple or list of 1s and -1s.
        The spin bit string.

    Returns
    -------
    d : int.

    Examples
    --------
    >>> spin_to_decimal((-1, -1, 1))
    6

    """
    return boolean_to_decimal(spin_to_boolean(b))


def pubo_value(x, P):
    r"""pubo_value.

    Find the value of
    :math:`\sum_{i,...,j} P_{i...j} x_{i} ... x_{j}`

    Parameters
    ----------
    x : dict or iterable.
        Maps boolean variable indices to their boolean values, 0 or 1. Ie
        ``x[i]`` must be the boolean value of variable i.
    P : dict, qubovert.utils.PUBOMatrix, or qubovert.PUBO object.
        Maps tuples of boolean variables indices to the P value.

    Returns
    -------
    value : float.
        The value of the PUBO with the given assignment `x`. Ie

    Example
    -------
    >>> P = {(0, 0): 1, (0, 1): -1}
    >>> x = {0: 1, 1: 0}
    >>> pubo_value(x, P)
    1

    """
    return sum(v for k, v in P.items() if all(x[i] for i in k))


def qubo_value(x, Q):
    r"""qubo_value.

    Find the value of
    :math:`\sum_{i,j} Q_{ij} x_{i} x_{j}`

    Parameters
    ----------
    x : dict or iterable.
        Maps boolean variable indices to their boolean values, 0 or 1. Ie
        ``x[i]`` must be the boolean value of variable i.
    Q : dict or qubovert.utils.QUBOMatrix object.
        Maps tuples of boolean variables indices to the Q value.

    Returns
    -------
    value : float.
        The value of the QUBO with the given assignment `x`. Ie

        >>> sum(
                Q[(i, j)] * x[i] * x[j]
                for i in range(n) for j in range(n)
            )

    Example
    -------
    >>> Q = {(0, 0): 1, (0, 1): -1}
    >>> x = {0: 1, 1: 0}
    >>> qubo_value(x, Q)
    1

    """
    return pubo_value(x, Q)


def puso_value(z, H):
    r"""puso_value.

    Find the value of
        :math:`\sum_{i,...,j} H_{i...j} z_{i} ... z_{j}`.

    Parameters
    ----------
    z: dict or iterable.
        Maps variable labels to their values, 1 or -1. Ie z[i] must be the
        value of variable i.
    H : dict, qubovert.utils.PUSOMatrix, or qubovert.PUSO object.
        Maps spin labels to values.

    Returns
    -------
    value : float.
        The value of the PUSO with the given assignment `z`.

    Example
    -------
    >>> H = {(0, 1): -1, (0,): 1}
    >>> z = {0: -1, 1: 1}
    >>> puso_value(z, H)
    0

    """
    return sum(
        v * pow(-1, [z[i] for i in k].count(-1) % 2)
        for k, v in H.items()
    )


def quso_value(z, L):
    r"""quso_value.

    Find the value of
        :math:`\sum_{i,j} J_{ij} z_{i} z_{j} + \sum_{i} h_{i} z_{i}`.

    The J's are encoded by keys with pairs of labels in L, and the h's are
    encoded by keys with a single label in L.

    Parameters
    ----------
    z: dict or iterable.
        Maps variable labels to their values, 1 or -1. Ie z[i] must be the
        value of variable i.
    L : dict, qubovert.utils.QUSOMatrix, or qubovert.QUSO object.
        Maps spin labels to values.

    Returns
    -------
    value : float.
        The value of the QUSO with the given assignment `z`.

    Example
    -------
    >>> L = {(0, 1): -1, (0,): 1}
    >>> z = {0: -1, 1: 1}
    >>> quso_value(z, L)
    0

    """
    return puso_value(z, L)


def _solve_bruteforce(D, all_solutions, valid, spin):
    """_solve_bruteforce.

    Helper function for solve_pubo_bruteforce and solve_puso_bruteforce.

    Iterate through all the possible solutions to a BO formulated problem
    and find the best one (the one that gives the minimum objective value). Do
    not use for large problem sizes! This is meant only for testing very small
    problems.

    Parameters
    ----------
    D : dict.
    all_solutions : boolean (optional, defaults to False).
        If all_solutions is set to True, all the best solutions to the problem
        will be returned rather than just one of the best. If the problem is
        very big, then it is best if ``all_solutions`` is False, otherwise this
        function will use a lot of memory.
    valid : function.
        ``valid`` takes in a bitstring or spinstring and outputs a boolean
        indicating whether that bitstring or spinstring is a valid solutions.
    spin : bool.
        Whether we're bruteforce solving a spin model or boolean model.

    Returns
    -------
    res : tuple (objective, solution).

        if all_solutions is False:
            objective : float.
                The best value of the problem.
            solution : dict.
                Maps the binary variable label to its solution value,
                {0, 1} if not spin else {-1, 1}.

        if all_solutions is True:
            objective : float.
                The best value of the problem
            solution : list of dicts.
                Each dictionary maps the label to the value of each binary
                variable. Ie each ``s`` in ``solution`` is a solution that
                gives the best objective function value.

    """
    if not D:
        return 0, ({} if not all_solutions else [{}])
    elif () in D:
        offset = D.pop(())
        if not D:
            D[()] = offset
            return offset, ({} if not all_solutions else [{}])
        D[()] = offset

    # if D is a Matrix object or QUBO, PUBO, etc, then these are defined
    try:
        N = D.num_binary_variables
        # could do D.reverse_mapping, but that creates a copy. We just need to
        # not mutate it here, then we don't have to waste time copying.
        mapping = D._reverse_mapping
    except AttributeError:
        var = set()
        for x in D:
            var.update(set(x))
        N = len(var)

        # map qubit name to 0 through N-1
        mapping = dict(enumerate(var))

    best = None, {}
    all_sols = {None: [{}]}
    convert = decimal_to_spin if spin else decimal_to_boolean

    for n in range(1 << N):
        test_sol = convert(n, N)
        x = {mapping[i]: v for i, v in enumerate(test_sol)}
        if not valid(x):
            continue
        v = puso_value(x, D) if spin else pubo_value(x, D)
        if all_solutions and (best[0] is None or v <= best[0]):
            best = v, x
            all_sols.setdefault(v, []).append(x)
        elif best[0] is None or v < best[0]:
            best = v, x

    if all_solutions:
        best = best[0], all_sols[best[0]]

    return best


def solve_pubo_bruteforce(P, all_solutions=False, valid=lambda x: True):
    """solve_pubo_bruteforce.

    Iterate through all the possible solutions to a PUBO formulated problem
    and find the best one (the one that gives the minimum objective value). Do
    not use for large problem sizes! This is meant only for testing very small
    problems.

    Parameters
    ----------
    P : dict, qubovert.utils.PUBOMatrix, or qubovert.PUBO object.
        Maps boolean variables labels to the P value.
    all_solutions : boolean (optional, defaults to False).
        If all_solutions is set to True, all the best solutions to the PUBO
        will be returned rather than just one of the best. If the problem is
        very big, then it is best if ``all_solutions`` is False, otherwise this
        function will use a lot of memory.
    valid : function (optional, defaults to ``lambda x: True``).
        ``valid`` takes in a bitstring and outputs a boolean
        indicating whether that bitstring is a valid solutions.

    Returns
    -------
    res : tuple (objective, solution).

        if all_solutions is False:
            objective : float.
                The best value of the PUBO.
            solution : dict.
                Maps the boolean variable label to its solution value, {0, 1}.

        if all_solutions is True:
            objective : float.
                The best value of the PUBO.
            solution : list of dicts.
                Each dictionary maps the label to the value of each boolean
                variable. Ie each ``s`` in ``solution`` is a solution that
                gives the best objective function value.

    Example
    -------
    To find the minimum of the problem
    :math:`obj = x_0x_1 + x_1x_2 - x_1 - 2x_2`,
    the P dictionary is

    >>> P = {(0, 1): 1, (1, 2): 1, (1,): -1, (2,): -2}

    Then to solve this P, run

    >>> obj_val, solution = solve_pubo_bruteforce(Q)
    >>> obj_val
    -2
    >>> solution
    {0: 0, 1: 0, 2: 1}

    ``obj_val`` will be the smallest value of ``obj``.
    ``solution`` will be a dictionary that indicates what each of :math:`x_0`,
    :math:`x_1`, and :math:`x_2` are for the solution. In this case,
    ``x = {0: 0, 1: 0, 2: 1}``, indicating that :math:`x_0` is 0, :math:`x_1`
    is 0, :math:`x_2` is 1.

    """
    return _solve_bruteforce(P, all_solutions, valid, False)


def solve_qubo_bruteforce(Q, all_solutions=False, valid=lambda x: True):
    """solve_qubo_bruteforce.

    Iterate through all the possible solutions to a QUBO formulated problem
    and find the best one (the one that gives the minimum objective value). Do
    not use for large problem sizes! This is meant only for testing very small
    problems.

    Parameters
    ----------
    Q : dict, qubovert.utils.QUBOMatrix, or qubovert.QUBO object.
        Maps boolean variables labels to the Q value.
    all_solutions : boolean (optional, defaults to False).
        If all_solutions is set to True, all the best solutions to the QUBO
        will be returned rather than just one of the best. If the problem is
        very big, then it is best if ``all_solutions`` is False, otherwise this
        function will use a lot of memory.
    valid : function (optional, defaults to ``lambda x: True``).
        ``valid`` takes in a bitstring and outputs a boolean
        indicating whether that bitstring is a valid solutions.

    Returns
    -------
    res : tuple (objective, solution).

        if all_solutions is False:
            objective : float.
                The best value of the QUBO.
            solution : dict.
                Maps the boolean variable label to its solution value, {0, 1}.

        if all_solutions is True:
            objective : float.
                The best value of the QUBO. Equal to
                ``sum(Q[(i, j)] * solution[x][i] * solution[x][j]) + offset``
                where `solution[x]` is one of the solutions to the QUBO.
            solution : list of dicts.
                Each dictionary maps the label to the value of each boolean
                variable. Ie each ``s`` in ``solution`` is a solution that
                gives the best objective function value.

    Example
    -------
    To find the minimum of the problem
    :math:`obj = x_0x_1 + x_1x_2 - x_1 - 2x_2`,
    the P dictionary is

    >>> Q = {(0, 1): 1, (1, 2): 1, (1,): -1, (2,): -2}

    Then to solve this Q, run

    >>> obj_val, solution = solve_qubo_bruteforce(Q)
    >>> obj_val
    -2
    >>> solution
    {0: 0, 1: 0, 2: 1}

    ``obj_val`` will be the smallest value of ``obj``.
    ``solution`` will be a dictionary that indicates what each of :math:`x_0`,
    :math:`x_1`, and :math:`x_2` are for the solution. In this case,
    ``x = {0: 0, 1: 0, 2: 1}``, indicating that :math:`x_0` is 0, :math:`x_1`
    is 0, :math:`x_2` is 1.

    """
    return solve_pubo_bruteforce(Q, all_solutions, valid)


def solve_puso_bruteforce(H, all_solutions=False, valid=lambda x: True):
    """solve_puso_bruteforce.

    Iterate through all the possible solutions to an PUSO formulated problem
    and find the best one (the one that gives the minimum objective value). Do
    not use for large problem sizes! This is meant only for testing very small
    problems.

    Parameters
    ----------
    H : dict, qubovert.utils.PUSOMatrix, or qubovert.PUSO object.
        Maps spin labels to values.
    all_solutions : boolean (optional, defaults to False).
        If all_solutions is set to True, all the best solutions to the QUSO
        will be returned rather than just one of the best. If the problem is
        very big, then it is best if ``all_solutions`` is False, otherwise this
        function will use a lot of memory.
    valid : function (optional, defaults to ``lambda x: True``).
        ``valid`` takes in a spinstring and outputs a boolean
        indicating whether that spinstring is a valid solutions.

    Returns
    -------
    res : tuple (objective, solution).

        if ``all_solutions`` is False:
            objective : float.
                The best value of the PUSO.
            solution : dict.
                Maps the spin variable label to its solution value, {-1, 1}.

        if ``all_solutions`` is True:
            objective : float.
                The best value of the PUSO.
            solution : list of dicts.
                Each dictionary maps the label to the value of each variable.
                Ie each ``s`` in ``solution`` is a solution that gives the best
                objective function value.

    Example
    -------
    To find the minimum of the problem
    :math:`obj = z_0z_1 + z_1z_2 - z_1 - 2z_2`,
    we have

    >>> H = {(0, 1): 1, (1, 2): 1, (1,): -1, (2,): -2}

    Then to solve this, run

    >>> obj_val, solution = solve_puso_bruteforce(H)
    >>> obj_val
    -3
    >>> solution
    {0: 1, 1: -1, 2: 1}

    ``obj_val`` will be the smallest value of ``obj``.
    ``solution`` will be a dictionary that indicates what each of
    :math:`z_0, z_1`, and :math:`z_2` are for the solution. In this case,
    ``z = {0: 1, 1: -1, 2: 1}``, indicating that :math:`z_0` is 1, :math:`z_1`
    is -1, :math:`z_2` is 1.

    """
    return _solve_bruteforce(H, all_solutions, valid, True)


def solve_quso_bruteforce(L, all_solutions=False, valid=lambda x: True):
    """solve_quso_bruteforce.

    Iterate through all the possible solutions to an QUSO formulated problem
    and find the best one (the one that gives the minimum objective value). Do
    not use for large problem sizes! This is meant only for testing very small
    problems.

    Parameters
    ----------
    L : dict, qubovert.utils.QUSOMatrix, or qubovert.QUSO object.
        Maps spin labels to values.
    all_solutions : boolean (optional, defaults to False).
        If all_solutions is set to True, all the best solutions to the QUSO
        will be returned rather than just one of the best. If the problem is
        very big, then it is best if ``all_solutions`` is False, otherwise this
        function will use a lot of memory.
    valid : function (optional, defaults to ``lambda x: True``).
        ``valid`` takes in a spinstring and outputs a boolean
        indicating whether that spinstring is a valid solutions.

    Returns
    -------
    res : tuple (objective, solution).

        if ``all_solutions`` is False:
            objective : float.
                The best value of the QUSO.
            solution : dict.
                Maps the spin variable label to its solution value, {-1, 1}.

        if ``all_solutions`` is True:
            objective : float.
                The best value of the PUSO.
            solution : list of dicts.
                Each dictionary maps the label to the value of each variable.
                Ie each ``s`` in ``solution`` is a solution that gives the best
                objective function value.

    Example
    -------
    To find the minimum of the problem
    :math:`obj = z_0z_1 + z_1z_2 - z_1 - 2z_2`,
    we have

    >>> L = {(0, 1): 1, (1, 2): 1, (1,): -1, (2,): -2}

    Then to solve this, run

    >>> obj_val, solution = solve_quso_bruteforce(L)
    >>> obj_val
    -3
    >>> solution
    {0: 1, 1: -1, 2: 1}

    ``obj_val`` will be the smallest value of ``obj``.
    ``solution`` will be a dictionary that indicates what each of
    :math:`z_0, z_1`, and :math:`z_2` are for the solution. In this case,
    ``z = {0: 1, 1: -1, 2: 1}``, indicating that :math:`z_0` is 1, :math:`z_1`
    is -1, :math:`z_2` is 1.

    """
    return solve_puso_bruteforce(L, all_solutions, valid)
